#!/usr/bin/env python

'''HappyNSM implements a node state manager with diagnostic capabilities'''
__revision__ = '$Revision$'

from os import environ, getuid, fork, close
from pwd import getpwuid
from signal import signal, SIGINT
from select import select
from syslog import syslog, LOG_INFO, LOG_ERR
from cPickle import dumps, loads
from elementtree.ElementTree import Element, SubElement, tostring, XML
from time import strftime, localtime
import re

from sss.server import BufferedPipe, MsgError, EventReceiver
from sss.ssslib import comm_lib
from sss.lrs import LRDataSet as DataSet, LRData as Data

class NodeState(Data):
    '''NodeStates track client states'''
    def UpdateNST(self, arglist):
        for test in self.element:
            if test.find("test") != None and test.find("test").text == arglist["test"]:
                self.element.remove(test)
            ne = Element("testresult")
            for field in ["test", "status", "output"]:
                SubElement(ne, field)
                ne.find(field).text = arglist[field]
            self.element.append(ne)
        if self.element.find("adminstate") == None:
            SubElement(self.element, "adminstate")
        self.element.find("adminstate").text = "online"
        for test in self.element.findall("testresult"):
            if test.find("status") != None and test.find("status").text != "0":
                self.element.find("adminstate").text = "offline"
        return

class happynsm(EventReceiver):
    '''Diagnostic node state manager'''
    __component__ = 'node-state-manager'
    __implementation__ = 'happynsm'
    __statefields__ = ['nsdata']
    __dispatch__ = {   'get-node-state':'nsdata.Get',
                       'set-node-state':'XSet',
                       'add-node-state':'nsdata.Add',
                       'del-node-state':'nsdata.Del',
                       'update-node-state':'nsdata.UpdateNST',
                       'diagnose':'Diagnose',
                       'events':'HandleEvent'}
    __validate__ = 1

    def __setup__(self):
        self.__subscriptions__ = [('process-manager', 'process_end', '**')]
        self.pgids = []
        self.nsdata = DataSet('node-states', 'node-state', NodeState, None, False)
        self.readConf('/etc/happynsm.conf')

    def readConf(self, filename):
        '''Read Configuration File'''
        try:
            for data in open(filename, 'r').readlines():
                (var, value) = data.split('=')
                value = value.strip()
                setattr(self, var, value.split(','))
        except IOError, e:
            print "Unable to open the happynsm.conf file: ", e

    def XSet(self, xml, (peer, port)):
        self.changed_nodes = {}
        result = self.nsdata.Get(xml, (peer, port), self.Set)
        self.evt.WriteMessage(dumps(self.changed_nodes.keys()))
        return result

    def Set(self, item, attrib):
        del attrib["node-state"]
        for x in attrib.keys():
            if item.element.find(x) != None and item.element.find(x).text != attrib[x]:
                item.element.find(x).text = attrib[x]
                if not (item.element.find('node').text in self.shanged_nodes):
                    self.changed_nodes[item.element.find('node').text] = 1

    def HandleEvent(self, xml, (peer, port)):
        '''process events from the event manager'''
        for event in xml.findall('event'):
            (c, m, d) = tuple([event.find(field).text for field in ['component', 'msg', 'data']])
            if c == 'process-manager':
                if m == 'process_end':
                    for (x, y, (peer, port)) in self.pgids:
                        if d == x:
                            self.recvDiagnosis(x, y, (peer, port))
                            self.pgids.remove((x, y, (peer, port)))
        return Element('event-ok')

    def Diagnose(self, xml, (peer, port)):
        '''run diagnostics on nodes'''
        if xml.find('node') != None:
            nodes = xml.find('node').text.strip()
        else:
            nodes = ''
        nodelist = nodes.split(',')
        if xml.find('test') != None:
            tests = xml.find('test').text.strip()
        else:
            tests = ''
        testlist = tests.split(',')
        #make this more advanced, support better description of nodes
        #if there is a nodename duplicated

        for testname in testlist:
            for executable in getattr(self, testname):
                self.sendMPDMessage(executable, nodelist, (peer, port)) 
        return Element('event-ok')
   
    def recvDiagnosis(self, pgid, executable, (peer, port)):
        test = executable.split("/")[len(executable.split("/"))-1].split(" ")[0].strip()
        comm = comm_lib(debug='-d' in argv)
        msg = XML("<wait-process-group><process-group><pgid>%s</pgid><submitter match='false'/><output match='false'/><exit-status><exit-code><host match='false'/><pid match='false'/><rank match='false'/><status match='false'/></exit-code></exit-status></process-group></wait-process-group>"%(pgid))
        pm = comm.ClientInit('process-manager')
        comm.SendMessage(pm, tostring(msg))
        ack = comm.RecvMessage(pm)
        comm.ClientClose(pm)
        try:
            xmlack = XML(ack)
            xmlexitcodes =  xmlack.find("process-group").find("exit-status").findall("exit-code")
            output = xmlack.find("process-group").findtext("output")
            syslog(LOG_INFO, "Got exit-info for pgid %s recieved successfully."%(pgid))
        except:
            print "there was an error reciving exit-codes from the Process Manager"
            return "<event-failed />"

        listoutput =  re.split('(\d+:)', output)
        del listoutput[0]
        corroutput = zip([listoutput[x] for x in range(0, len(listoutput), 2)],
                         [listoutput[x] for x in range(1, len(listoutput), 2)])

        noderes = {}
        for node in xmlexitcodes:
            outputlines = ""
            for (x, y) in corroutput:
                if node.find("rank") != None and node.find("rank").text + ":" == x:
                    outputlines += y + "\n"

            noderes[node] = Element("testresult")
            SubElement(noderes[node], "node")
            noderes[node].find("node").text = node.find("host").text
            SubElement(noderes[node], "test")
            noderes[node].find("test").text = test
            SubElement(noderes[node], "status")
            noderes[node].find("status").text = node.find("status").text
            SubElement(noderes[node], "output")
            noderes[node].find("output").text = outputlines

        xml = Element("get-node-state")
        for node in noderes.keys():
            ne = Element("node-state")
            SubElement(ne, "adminstate")
            ne.find("adminstate").attrib['match'] = "false"
            SubElement(ne, "node")
            ne.find("node").text = node.find("host").text
            SubElement(ne, "state")
            ne.find("state").attrib['match'] = "false"
            xml.append(ne)
        nodestates = self.nsdata.Get(xml, (peer, port))
        for node in noderes.keys():
            try:
                nodestate = [nds for nds in nodestates.findall('node-state') if nds.find('node').text == noderes[node].find('node').text]
            except:
                syslog(LOG_ERR, "Failed to get node-state event")
                return
             
        if nodestate != []:
            self.nsdata.Get(XML('''<update-node-state><test>%s</test><status>%s</status><output>%s</output><node-state><adminstate match='false'/><node>%s</node><state match='false'/></node-state></update-node-state>''' % (noderes[node].find("test").text, noderes[node].find("status").text, noderes[node].find("output").text, noderes[node].find("node").text)), (peer, port), lambda x, y:x.UpdateNST(y))

        else:
            syslog(LOG_ERR, "Failed to send state-changed event")
        return

    def sendMPDMessage(self, executable, nodelist, (peer, port)):
        #Simply, this creates, sends, and gets output from the Process Manager. 
        numprocs   = str(len(nodelist))
        submitter = getpwuid(getuid())[0]

        path = environ['PATH']
        cwd = environ['PWD']
        msg = Element("create-process-group")
        SubElement(msg, "pgid")
        msg.find("pgid").attrib['match'] = 'false'
        SubElement(msg, "submitter")
        msg.find("submitter").text = submitter
        SubElement(msg, "totalprocs")
        msg.find("totalprocs").text = numprocs
        ps = SubElement(msg, "process-spec")
        SubElement(ps, "user")
        ps.find("user").text = submitter
        SubElement(ps, "path")
        ps.find("path").text = path
        SubElement(ps, "cwd")
        ps.find("cwd").text = cwd
        SubElement(ps, "exec")
        ps.find("exec").text = executable
        #host spec
        nodetext = ''
        for node in nodelist:
            nodetext = nodetext + node + '\n'
        hs = SubElement(msg, "host-spec")
        #SubElement(hs, "check")
        #hs.find("check").text = 'yes'
        hs.text = nodetext

        comm = comm_lib(debug='-d' in argv)
        process_manager = comm.ClientInit('process-manager')
        comm.SendMessage(process_manager, tostring(msg))
        ack = comm.RecvMessage(process_manager)
        comm.ClientClose(process_manager)
        r = XML(ack)
        pgid = ''
        if r.find("pgid") != None:
            pgid = r.find('pgid').text
            self.pgids.append((pgid, executable, (peer, port)))
        #add this pgid to acceptable pgids list
        if not pgid:
            syslog(LOG_ERR, "Failed to recieve PGID")
            return
        else:
            syslog(LOG_INFO, "Got PGID %s\nwaiting for event data" % (pgid))
        return
   
    def ChildSigHand(self, signum, frame):
        raise SystemExit, 0
    
    def __setupChildren__(self):
        self.evt = BufferedPipe()
        self.pipes.append(self.evt)
        pid = fork()
        if pid == 0:
            signal(SIGINT, self.ChildSigHand)
            self.readsockets = [self.evt.r]
            self.socketdispatch = {self.evt.r:self.ProcessChanges_cb}
            self.ServeForever()
        else:
            close(self.evt.r)
            return [pid]

    def ProcessChanges_cb(self, pipe, inl=None):
        '''Read changes from pipe and create events for them'''
        while 1:
            try:
                (infd, out, x) = select([pipe], [], [], 5)
                # Catch and disregard the traceback on child process exit
            except:
                pass
            if infd != []:
                self.evt.ReadMessages()

            while self.evt.buffer:
                try:
                    changed_nodes = loads(self.evt.ReadMessage())
                except MsgError:
                    break
      
                # Send state changed message
                doc = Element('add-event')
                for node in changed_nodes:
                    docse = SubElement(doc, 'event')
                    SubElement(docse, "component")
                    docse.find("component").text = self.__component__
                    SubElement(docse, "msg")
                    docse.find("msg").text = 'state-changed'
                    SubElement(docse, "data")
                    docse.find("data").text = node
                    SubElement(docse, "time")
                    docse.find("time").text = strftime("%c",localtime())
                if len(doc) > 0:
                    try:
                        c = comm_lib()
                        h = c.ClientInit('event-manager')
                        c.SendMessage(h, tostring(doc))
                        c.RecvMessage(h)
                        c.ClientClose(h)
                    except:
                        syslog(LOG_ERR, "Failed to send state-changed event")
      
                del doc

if __name__ == '__main__':
    from getopt import getopt, GetoptError
    from sys import argv
    try:
        (opts, args) = getopt(argv[1:], 'd', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nnsm.py [-d] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [item for item in opts if item[0] == '--daemon']
    ldebug = len([item for item in opts if item[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    server = happynsm(debug=ldebug)
    server.ServeForever()
